
/******************************************************************************
 *
 *  This file is part of canu, a software program that assembles whole-genome
 *  sequencing reads into contigs.
 *
 *  This software is based on:
 *    'Celera Assembler' r4587 (http://wgs-assembler.sourceforge.net)
 *    the 'kmer package' r1994 (http://kmer.sourceforge.net)
 *
 *  Except as indicated otherwise, this is a 'United States Government Work',
 *  and is released in the public domain.
 *
 *  File 'README.licenses' in the root directory of this distribution
 *  contains full conditions and disclaimers.
 */

#ifndef BINOMIAL_BOUND_H
#define BINOMIAL_BOUND_H

#include "types.H"

int
Binomial_Bound(int e, double p, int Start);

void
Initialize_Match_Limit(int32 *ml, double maxErate, int32 MAX_ERRORS);

#endif
