
/******************************************************************************
 *
 *  This file is part of seqrequester, a tool for summarizing, extracting,
 *  generating and modifying DNA sequences.
 *
 *  This software is based on:
 *    'Canu' v2.0              (https://github.com/marbl/canu)
 *  which is based on:
 *    'Celera Assembler' r4587 (http://wgs-assembler.sourceforge.net)
 *    the 'kmer package' r1994 (http://kmer.sourceforge.net)
 *
 *  Except as indicated otherwise, this is a 'United States Government Work',
 *  and is released in the public domain.
 *
 *  File 'README.licenses' in the root directory of this distribution
 *  contains full conditions and disclaimers.
 */

#ifndef SEQREQUESTER_INCLUDE_H
#error  Include simulate.H only from seqrequester.H
#endif


class microsatelliteParameters {
public:
  microsatelliteParameters() {
  }
  ~microsatelliteParameters() {
  }

  void      initialize(void) {
  }

  void      finalize(void) {
  }


  //char const  *seqName;   vector<inputs>
  char const  *outPrefix  = nullptr;

  uint32       window     = 128;

  bool         report_ga  = false;
  bool         report_gc  = false;
  bool         report_at  = false;

  bool         report_legacy = false;


#if 0
  char         seqName[FILENAME_MAX+1];
  char         distribName[FILENAME_MAX+1];
  char         outputName[FILENAME_MAX+1];
#endif
};
