//  Automagically generated by version_update.pl!  Do not commit!
#define CANU_VERSION_LABEL     "release"
#define CANU_VERSION_MAJOR     "2"
#define CANU_VERSION_MINOR     "2"
#define CANU_VERSION_COMMITS   ""
#define CANU_VERSION_REVISION  "0"
#define CANU_VERSION_HASH      ""

#define CANU_VERSION           "canu 2.2\n"

#undef  MERYL_UTILITY_VERSION
#define MERYL_UTILITY_VERSION CANU_VERSION
