/* config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Define if building universal (internal helper macro) */
/* #undef AC_APPLE_UNIVERSAL_BUILD */

/* Define to 1 if you have the `aligned_alloc' function. */
#define HAVE_ALIGNED_ALLOC 1

/* define to 1 if the C compiler supports ALTIVEC */
#define HAVE_ALTIVEC 0

/* define to 1 if the C compiler supports AVX2 */
#define HAVE_AVX2 1

/* define to 1 if the C compiler supports AVX2 _mm256_extract_epi16 */
#define HAVE_AVX2_MM256_EXTRACT_EPI16 1

/* define to 1 if the C compiler supports AVX2 _mm256_extract_epi32 */
#define HAVE_AVX2_MM256_EXTRACT_EPI32 1

/* define to 1 if the C compiler supports AVX2 _mm256_extract_epi64 */
#define HAVE_AVX2_MM256_EXTRACT_EPI64 1

/* define to 1 if the C compiler supports AVX2 _mm256_extract_epi8 */
#define HAVE_AVX2_MM256_EXTRACT_EPI8 1

/* define to 1 if the C compiler supports AVX2 _mm256_insert_epi16 */
#define HAVE_AVX2_MM256_INSERT_EPI16 1

/* define to 1 if the C compiler supports AVX2 _mm256_insert_epi32 */
#define HAVE_AVX2_MM256_INSERT_EPI32 1

/* define to 1 if the C compiler supports AVX2 _mm256_insert_epi64 */
#define HAVE_AVX2_MM256_INSERT_EPI64 1

/* define to 1 if the C compiler supports AVX2 _mm256_insert_epi8 */
#define HAVE_AVX2_MM256_INSERT_EPI8 1

/* define to 1 if the C compiler supports AVX2 _mm256_set1_epi64x */
#define HAVE_AVX2_MM256_SET1_EPI64X 1

/* define to 1 if the C compiler supports AVX2 _mm256_set_epi64x */
#define HAVE_AVX2_MM256_SET_EPI64X 1

/* define to 1 if the C compiler supports AVX512BW */
#define HAVE_AVX512BW 1

/* define to 1 if the C compiler supports AVX512F */
#define HAVE_AVX512F 1

/* define to 1 if the C compiler supports AVX512VBMI */
#define HAVE_AVX512VBMI 1

/* define to 1 if using clock_gettime CLOCK_MONOTONIC */
#define HAVE_CLOCK_GETTIME_MONOTONIC 0

/* define to 1 if using clock_gettime CLOCK_REALTIME */
#define HAVE_CLOCK_GETTIME_REALTIME 0

/* define to 1 if using clock_get_time */
#define HAVE_CLOCK_GET_TIME 0

/* Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Define to 1 if you have the `filelength' function. */
/* #undef HAVE_FILELENGTH */

/* Define to 1 if you have the `getopt' function. */
#define HAVE_GETOPT 1

/* define to 1 if using windows GetSystemTimeAsFileTime */
#define HAVE_GETSYSTEMTIMEASFILETIME 0

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the <malloc.h> header file. */
/* #undef HAVE_MALLOC_H */

/* Define to 1 if you have the `memalign' function. */
/* #undef HAVE_MEMALIGN */

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the `memset' function. */
#define HAVE_MEMSET 1

/* define to 1 if the C compiler supports NEON */
#define HAVE_NEON 0

/* Define if OpenMP is enabled */
#define HAVE_OPENMP 1

/* Define to 1 if you have the `poll' function. */
#define HAVE_POLL 1

/* Define to 1 if you have the <poll.h> header file. */
#define HAVE_POLL_H 1

/* Define to 1 if you have the `posix_memalign' function. */
#define HAVE_POSIX_MEMALIGN 1

/* Define to 1 if you have the `SetUnhandledExceptionFilter' function. */
/* #undef HAVE_SETUNHANDLEDEXCEPTIONFILTER */

/* define to 1 if the C compiler supports SSE2 */
#define HAVE_SSE2 1

/* define to 1 if the C compiler supports SSE2 _mm_set1_epi64x */
#define HAVE_SSE2_MM_SET1_EPI64X 1

/* define to 1 if the C compiler supports SSE2 _mm_set_epi64x */
#define HAVE_SSE2_MM_SET_EPI64X 1

/* define to 1 if the C compiler supports SSE4.1 */
#define HAVE_SSE41 1

/* define to 1 if the C compiler supports SSE4.1 _mm_extract_epi64 */
#define HAVE_SSE41_MM_EXTRACT_EPI64 1

/* define to 1 if the C compiler supports SSE4.1 _mm_insert_epi64 */
#define HAVE_SSE41_MM_INSERT_EPI64 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if the system has the type `struct __stat64'. */
/* #undef HAVE_STRUCT___STAT64 */

/* Define to 1 if you have the <sys/mman.h> header file. */
#define HAVE_SYS_MMAN_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* define to 1 if the CXX compiler supports variadic macros */
#define HAVE_VARIADIC_MACROS 1

/* Define to 1 if you have the <windows.h> header file. */
/* #undef HAVE_WINDOWS_H */

/* define to 1 if the C compiler supports xgetbv */
#define HAVE_XGETBV 1

/* define to 1 if zlib is available */
#define HAVE_ZLIB 1

/* Define to 1 if you have the `_aligned_malloc' function. */
/* #undef HAVE__ALIGNED_MALLOC */

/* Define to 1 if you have the `_fstat64' function. */
/* #undef HAVE__FSTAT64 */

/* Define to 1 if you have the `_stat64' function. */
/* #undef HAVE__STAT64 */

/* int64_t suffix is i64 */
/* #undef INT64_LITERAL_SUFFIX_I64 */

/* int64_t suffix is LL */
#define INT64_LITERAL_SUFFIX_LL 1

/* int64_t suffix is unknown */
/* #undef INT64_LITERAL_SUFFIX_UNKNOWN */

/* Define to the sub-directory where libtool stores uninstalled libraries. */
#define LT_OBJDIR ".libs/"

/* Name of package */
#define PACKAGE "parasail"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "jeffrey.daily@gmail.com"

/* Define to the full name of this package. */
#define PACKAGE_NAME "Pairwise Sequence Alignment Library"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "Pairwise Sequence Alignment Library 2.4.2"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "parasail"

/* Define to the home page for this package. */
#define PACKAGE_URL ""

/* Define to the version of this package. */
#define PACKAGE_VERSION "2.4.2"

/* The size of `int', as computed by sizeof. */
#define SIZEOF_INT 4

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* Version number of package */
#define VERSION "2.4.2"

/* Define WORDS_BIGENDIAN to 1 if your processor stores words with the most
   significant byte first (like Motorola and SPARC, unlike Intel). */
#if defined AC_APPLE_UNIVERSAL_BUILD
# if defined __BIG_ENDIAN__
#  define WORDS_BIGENDIAN 1
# endif
#else
# ifndef WORDS_BIGENDIAN
/* #  undef WORDS_BIGENDIAN */
# endif
#endif

/* Enable large inode numbers on Mac OS X 10.5.  */
#ifndef _DARWIN_USE_64_BIT_INODE
# define _DARWIN_USE_64_BIT_INODE 1
#endif

/* Number of bits in a file offset, on hosts where this is settable. */
/* #undef _FILE_OFFSET_BITS */

/* Define for large files, on AIX-style hosts. */
/* #undef _LARGE_FILES */

/* Define for Solaris 2.5.1 so the uint32_t typedef from <sys/synch.h>,
   <pthread.h>, or <semaphore.h> is not used. If the typedef were allowed, the
   #define below would cause a syntax error. */
/* #undef _UINT32_T */

/* Define for Solaris 2.5.1 so the uint64_t typedef from <sys/synch.h>,
   <pthread.h>, or <semaphore.h> is not used. If the typedef were allowed, the
   #define below would cause a syntax error. */
/* #undef _UINT64_T */

/* Define for Solaris 2.5.1 so the uint8_t typedef from <sys/synch.h>,
   <pthread.h>, or <semaphore.h> is not used. If the typedef were allowed, the
   #define below would cause a syntax error. */
/* #undef _UINT8_T */

/* Define to empty if `const' does not conform to ANSI C. */
/* #undef const */

/* Define to `__inline__' or `__inline' if that's what the C compiler
   calls it, or to nothing if 'inline' is not supported under any name.  */
#ifndef __cplusplus
/* #undef inline */
#endif

/* Define to the type of a signed integer type of width exactly 16 bits if
   such a type exists and the standard includes do not define it. */
/* #undef int16_t */

/* Define to the type of a signed integer type of width exactly 32 bits if
   such a type exists and the standard includes do not define it. */
/* #undef int32_t */

/* Define to the type of a signed integer type of width exactly 64 bits if
   such a type exists and the standard includes do not define it. */
/* #undef int64_t */

/* Define to the type of a signed integer type of width exactly 8 bits if such
   a type exists and the standard includes do not define it. */
/* #undef int8_t */

/* Define to `long int' if <sys/types.h> does not define. */
/* #undef off_t */

/* Define to the equivalent of the C99 'restrict' keyword, or to
   nothing if this is not supported.  Do not define if restrict is
   supported directly.  */
#define restrict __restrict
/* Work around a bug in Sun C++: it does not support _Restrict or
   __restrict__, even though the corresponding Sun C compiler ends up with
   "#define restrict _Restrict" or "#define restrict __restrict__" in the
   previous line.  Perhaps some future version of Sun C++ will work with
   restrict; if so, hopefully it defines __RESTRICT like Sun C does.  */
#if defined __SUNPRO_CC && !defined __RESTRICT
# define _Restrict
# define __restrict__
#endif

/* Define to `unsigned int' if <sys/types.h> does not define. */
/* #undef size_t */

/* Define to the type of an unsigned integer type of width exactly 16 bits if
   such a type exists and the standard includes do not define it. */
/* #undef uint16_t */

/* Define to the type of an unsigned integer type of width exactly 32 bits if
   such a type exists and the standard includes do not define it. */
/* #undef uint32_t */

/* Define to the type of an unsigned integer type of width exactly 64 bits if
   such a type exists and the standard includes do not define it. */
/* #undef uint64_t */

/* Define to the type of an unsigned integer type of width exactly 8 bits if
   such a type exists and the standard includes do not define it. */
/* #undef uint8_t */
